#!/usr/bin/env python
# -*- coding:utf-8 -*-
import time
import uuid
import collections
from bson import ObjectId
from faker import Faker
from faker_config import faker_config


# 为了减少依赖，从Thrift中拷贝过来
class TType(object):
    STOP = 0
    VOID = 1
    BOOL = 2
    BYTE = 3
    I08 = 3
    DOUBLE = 4
    I16 = 6
    I32 = 8
    I64 = 10
    STRING = 11
    UTF7 = 11
    STRUCT = 12
    MAP = 13
    SET = 14
    LIST = 15
    UTF8 = 16
    UTF16 = 17

fake = Faker()
LIST_COUNT = 2
MAP_COUNT = 2


INT_CONFIG = collections.OrderedDict()
UNICODE_CONFIG = collections.OrderedDict()
for fake_function, param_names in faker_config.get('int').items():
    for param_name in param_names:
        INT_CONFIG[param_name] = fake_function
for fake_function, param_names in faker_config.get('unicode').items():
    for param_name in param_names:
        UNICODE_CONFIG[param_name] = fake_function


def fake_request_param(request_type):
    return _fake_struct(
        request_type.__name__, (request_type, request_type.thrift_spec))


def fake_param(attr_name, target_ttype, ttype_conf):
    fake_func = _DUMPS_FUNC[target_ttype]
    return fake_func(attr_name, ttype_conf)


def _fake_bool(attr_name, ttype_conf):
    return fake.boolean()


def _fake_int(attr_name, ttype_conf):
    ret = __special_int_fake(attr_name)
    if ret:
        return ret
    return fake.random_int(min=0, max=10)


def _fake_float(attr_name, ttype_conf):
    return fake.pyfloat(left_digits=2, right_digits=3)


def _fake_unicode(attr_name, ttype_conf):
    ret = __special_unicode_fake(attr_name)
    if ret:
        return ret
    return unicode(fake.pystr(max_chars=5))


def _fake_list(attr_name, ttype_conf):
    target_type = ttype_conf[0]
    thrift_spec = ttype_conf[1]
    return [fake_param(attr_name, target_type, thrift_spec) for i in range(LIST_COUNT)]


def _fake_set(attr_name, ttype_conf):
    return set(_fake_list(ttype_conf))


def _fake_map(attr_name, ttype_conf):
    key_ttype = ttype_conf[0]
    key_ttype_conf = ttype_conf[1]
    value_ttype = ttype_conf[2]
    value_ttype_conf = ttype_conf[3]
    ret = collections.OrderedDict()
    for i in range(MAP_COUNT):
        tkey = fake_param(attr_name, key_ttype, key_ttype_conf)
        tvalue = fake_param(attr_name, value_ttype, value_ttype_conf)
        ret[tkey] = tvalue
    return ret


def _fake_struct(attr_name, ttype_conf):
    ret = collections.OrderedDict()
    struct_spec = ttype_conf[1]
    for spec in struct_spec:
        if spec:
            attr_target_type = spec[1]
            attr_name = spec[2]
            attr_ttype_conf = spec[3]
            attr_default = spec[4]
            if attr_default is not None:
                value = attr_default
            else:
                value = fake_param(attr_name, attr_target_type, attr_ttype_conf)
            ret[attr_name] = value
    return ret


_DUMPS_FUNC = {
    TType.BOOL: _fake_bool,
    TType.BYTE: _fake_int,
    TType.I08: _fake_int,
    TType.I16: _fake_int,
    TType.I32: _fake_int,
    TType.I64: _fake_int,
    TType.DOUBLE: _fake_float,
    TType.STRING: _fake_unicode,
    TType.LIST: _fake_list,
    TType.SET: _fake_set,
    TType.MAP: _fake_map,
    TType.STRUCT: _fake_struct
}


def __special_int_fake(attr_name):
    for config_name, fake_func in INT_CONFIG.items():
        if config_name in attr_name:
            if fake_func == 'date_time_this_month':
                ret = getattr(fake, fake_func)(after_now=True)
                ret = int(time.mktime(ret.timetuple()))
            else:
                ret = getattr(fake, fake_func)()
            return ret


def __special_unicode_fake(attr_name):
    for config_name, fake_func in UNICODE_CONFIG.items():
        if config_name in attr_name:
            if fake_func == 'uuid':
                ret = str(uuid.uuid4())
            elif fake_func == 'object_id':
                ret = str(ObjectId())
            else:
                ret = getattr(fake, fake_func)()
            return ret



#!/usr/bin/env python
# -*- coding:utf-8 -*-
"""
    通过解析IDL自动生成Pytest单测框架,
    通过fake辅助生成部分参数,
    并且支持单测参数和单测逻辑的向后兼容;
"""
import os
import inspect
import collections
import json
import yaml
from thrift_fake import fake_request_param
from dict_format_print import format_dict
from dict_replace import replace_dict

TAP = '    '


def __to_underscore(str_item):
    return ''.join('_%s' % c if c.isupper() else c
                   for c in str_item).strip('_').lower()


def __get_need_test_methods_from_handler(test_handler):
    """从指定Handler获取需要测试的方法（包含一些非IDL的方法，后面会处理掉）"""
    need_test_methods = list()
    for name, obj in inspect.getmembers(test_handler):
        if inspect.ismethod(obj):
            need_test_methods.append(name)
    return need_test_methods


def __get_service_info_from_thrift(root_path, idl_service, need_test_methods):
    """从指定IDL_Service和request_config配置表中，获取测试方法和Request的映射表"""
    customized_request_config = yaml.load(
        open(os.path.join(root_path, 'red_test', 'request_config.yaml')))

    method_request = collections.OrderedDict()

    idl_method_request = dict([
        (name[:-5], obj) for name, obj in inspect.getmembers(idl_service)
        if inspect.isclass(obj) and '_args' in name])

    for method_name in need_test_methods:
        if customized_request_config and method_name in customized_request_config:
            method_request[method_name] = customized_request_config[method_name]
        elif method_name in idl_method_request:
            request_obj_name = idl_method_request[method_name].thrift_spec[2][3][0].__name__
            method_request[method_name] = {__to_underscore(request_obj_name): request_obj_name}

    return method_request


def __get_request_info_from_thrift(idl_request, method_request):
    """从指定IDL_Request中，获取测试Request名和Request类的映射表"""
    request_name_obj = collections.OrderedDict()

    request_name__obj_name = collections.OrderedDict()
    for request_config in method_request.values():
        request_name__obj_name.update(request_config)

    request_obj_name__obj = dict([
        (name, obj)for name, obj in inspect.getmembers(idl_request) if inspect.isclass(obj)])

    for name, obj_name in request_name__obj_name.items():
        request_name_obj[name] = request_obj_name__obj[obj_name]
    return request_name_obj


def __get_existed_requests_param(root_path):
    """获取已经写过的单测参数，以支持后面的复写，从而支持单测参数的向后兼容"""
    existed_requests_param = dict()
    if os.path.exists(os.path.join(
            root_path, 'red_test', 'handler', 'requests_param.py')):
        import red_test.handler.requests_param as existed_param
        for name, obj in inspect.getmembers(existed_param):
            if name != '__builtins__' and type(obj) == dict:
                existed_requests_param[name] = obj
    return existed_requests_param


def __get_existed_handler_methods(root_path):
    """获取已经写过的单测方法的逻辑，以支持后面的复写，从而支持单测逻辑的向后兼容"""
    existed_handler_methods = dict()
    if os.path.exists(os.path.join(
            root_path, 'red_test', 'handler', 'handler_methods.py')):
        import red_test.handler.handler_methods as existed_method
        with open(os.path.join(
                root_path, 'red_test', 'handler', 'handler_methods.py'), 'r') as f:
            whole_file = f.readlines()
            prev_name = ''
            for name, obj in inspect.getmembers(existed_method):
                if inspect.isfunction(obj):
                    first_line_no = obj.__code__.co_firstlineno
                    existed_handler_methods[name] = first_line_no
                    if prev_name:
                        prev_first_line_no = existed_handler_methods[prev_name]
                        existed_handler_methods[prev_name] = whole_file[
                            prev_first_line_no - 1: first_line_no - 1]
                    prev_name = name
            if prev_name:
                prev_first_line_no = existed_handler_methods[prev_name]
                existed_handler_methods[prev_name] = whole_file[
                    prev_first_line_no - 1:]
    return existed_handler_methods


def __write_test_first(root_path):
    """构建单测的第一层架构"""
    if not os.path.exists(os.path.join(root_path, 'red_test')):
        os.mkdir(os.path.join(root_path, 'red_test'))

        with open(os.path.join(
                root_path, 'red_test', '__init__.py'), 'w') as f:
            print 'Create root __init__ --------------------'
            contents = list()
            contents.append('#!/usr/bin/env python\n')
            contents.append('# -*- coding:utf-8 -*-\n')
            contents.append('import os\n')
            contents.append('import sys\n')
            contents.append('myPath = os.path.dirname(os.path.abspath(__file__))\n')
            contents.append('sys.path.insert(0, myPath + \'/../\')\n')
            f.writelines(contents)

        with open(os.path.join(
                root_path, 'red_test', 'conftest.py'), 'w') as f:
            print 'Create root conftest --------------------'
            contents = list()
            contents.append('#!/usr/bin/env python\n')
            contents.append('# -*- coding:utf-8 -*-\n')
            contents.append('# 创建单测环境\n')
            contents.append('import pytest\n')
            f.writelines(contents)

        with open(os.path.join(
                root_path, 'red_test', 'request_config.yaml'), 'w') as f:
            print 'Create root request_config --------------------'
            contents = list()
            contents.append('# 配置单测参数\n')
            f.writelines(contents)


def __write_test_second(root_path):
    """构建单测的第二层架构"""
    if not os.path.exists(os.path.join(root_path, 'red_test', 'handler')):
        os.mkdir(os.path.join(root_path, 'red_test', 'handler'))

        with open(os.path.join(
                root_path, 'red_test', 'handler', '__init__.py'), 'w') as f:
            print 'Create __init__ --------------------'
            contents = list()
            contents.append('#!/usr/bin/env python\n')
            contents.append('# -*- coding:utf-8 -*-\n')
            f.writelines(contents)

        with open(os.path.join(
                root_path, 'red_test', 'handler', 'conftest.py'), 'w') as f:
            print 'Create conftest --------------------'
            contents = list()
            contents.append('#!/usr/bin/env python\n')
            contents.append('# -*- coding:utf-8 -*-\n')
            contents.append('from redrpc import red_context\n')
            contents.append('from pytest_requests import *\n\n\n')
            contents.append('@pytest.fixture(scope=\'function\', autouse=True)\n')
            contents.append('def handler_context():\n')
            contents.append(TAP + 'return red_context.get().gen_thrift_context()\n')
            f.writelines(contents)


def __write_request_params(root_path, request_name_obj, existed_requests_param):
    """复写单测参数，支持向后兼容"""
    with open(os.path.join(
            root_path, 'red_test', 'handler', 'requests_param.py'), 'w') as f:
        print 'Create test params --------------------'
        requests_param = list()
        requests_param.append('#!/usr/bin/env python\n')
        requests_param.append('# -*- coding:utf-8 -*-\n')
        for name, obj in request_name_obj.items():
            param_name = name + '_params'
            request_param = fake_request_param(obj)
            if param_name in existed_requests_param:
                existed_request_param = existed_requests_param[param_name]
                replace_dict(existed_request_param, request_param)
                if existed_request_param != request_param:
                    print 'Change test params of ' + name
            else:
                print 'Create test params of ' + name

            request_param = json.dumps(request_param)
            request_param = request_param.replace('false', 'False')
            request_param = request_param.replace('true', 'True')
            requests_param.append(
                param_name + ' = ' + format_dict(request_param) + '\n\n')
        f.writelines(requests_param)


def __write_pytest_request(root_path, idl_request, request_name_obj):
    """复写pytest_request_fixture"""
    with open(os.path.join(
            root_path, 'red_test', 'handler', 'pytest_requests.py'), 'w') as f:
        pytest_requests = list()
        pytest_requests.append('#!/usr/bin/env python\n')
        pytest_requests.append('# -*- coding:utf-8 -*-\n')
        pytest_requests.append('import pytest\n')
        pytest_requests.append('from doraemon.thrift import wrap_struct\n')
        pytest_requests.append('from red_test.handler.requests_param import *\n')
        pytest_requests.append('from ' + idl_request.__name__ + ' import *\n\n\n')
        for name, obj in request_name_obj.items():
            method_decorator = '@pytest.fixture(params=[' + name + '_params])\n'
            method_header = 'def ' + name + '(request):\n'
            method_content = TAP + 'request = request.params\n'
            method_content += TAP + 'return wrap_struct(request, ' + obj.__name__ + ')\n'
            pytest_requests.append(method_decorator + method_header + method_content + '\n\n')
        f.writelines(pytest_requests)


def __write_test_methods(root_path, method_request, existed_handler_methods):
    """复写单测逻辑，支持向后兼容"""
    with open(os.path.join(
            root_path, 'red_test', 'handler', 'handler_methods.py'), 'w') as f:
        print 'Create test methods --------------------'
        handler_methods = list()
        handler_methods.append('#!/usr/bin/env python\n')
        handler_methods.append('# -*- coding:utf-8 -*-\n')
        handler_methods.append('from handler import Handler\n')
        handler_methods.append('handler_object = Handler()\n\n\n')
        for method_name, request_config in method_request.items():
            test_method_name = 'test_' + method_name

            if test_method_name in existed_handler_methods:
                handler_methods.extend(existed_handler_methods[test_method_name])
            else:
                print 'Create test method: ' + method_name
                request_names = request_config.keys()
                method_header = (
                    'def ' + test_method_name + '(handler_context, '
                    + ', '.join(request_names) + '):\n')
                method_content = TAP + 'try: \n'
                method_content += TAP + TAP + 'result = handler_object.' + method_name + '(\n'
                method_content += TAP + TAP + TAP + 'handler_context,\n'
                method_content += TAP + TAP + TAP + request_names[0] + ').data\n'
                method_content += TAP + TAP + 'assert result is not None\n'
                method_content += TAP + 'except Exception as e:\n'
                method_content += TAP + TAP + 'print e.message\n'
                method_content += TAP + TAP + 'assert False\n'
                handler_methods.append(method_header + method_content + '\n\n')
        f.writelines(handler_methods)


def generate_red_test(root_path, test_handler, test_service, test_request):
    """自动生成Red Test"""
    __write_test_first(root_path)
    __write_test_second(root_path)

    need_test_methods = __get_need_test_methods_from_handler(test_handler)
    method_request = __get_service_info_from_thrift(root_path, test_service, need_test_methods)
    request_name_obj = __get_request_info_from_thrift(test_request, method_request)
    existed_requests_param = __get_existed_requests_param(root_path)
    existed_handler_methods = __get_existed_handler_methods(root_path)

    __write_request_params(root_path, request_name_obj, existed_requests_param)
    __write_pytest_request(root_path, test_request, request_name_obj)
    __write_test_methods(root_path, method_request, existed_handler_methods)


# if __name__ == '__main__':
#     测试案例
#     from handler.stockage_handler import StockageHandler
#     import stockage_service.request.ttypes as test_request
#     import stockage_service.StockageService as test_service
#     generate_red_test(ENV['root'], StockageHandler, test_service, test_request)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
TAP = "    "


def format_dict(dict_string):
    """格式化dict"""
    item_list = [item for item in dict_string]
    tap_deep = 0
    line_length = 0

    for i in range(len(item_list)):
        line_length += 1
        if line_length > 100:
            j = i
            while j > 0 and item_list[j] != ",":
                j -= 1
            add_return_after(item_list, j + 1, tap_deep)
            line_length = (tap_deep - 1) * 4 + (i - j)

        if item_list[i] == "{":
            tap_deep += 1
            add_return_before(item_list, i + 1, tap_deep)
            line_length = tap_deep * 4 + 1
        elif item_list[i] == "}":
            add_return_after(item_list, i - 1, tap_deep - 1)
            line_length = (tap_deep - 1) * 4 + 1
            tap_deep -= 1

    return ''.join(item_list)


def add_return_after(item_list, j, tap_deep):
    item_list[j] += "\n"
    for i in range(tap_deep):
        item_list[j] += TAP


def add_return_before(item_list, j, tap_deep):
    for i in range(tap_deep):
        item_list[j] = TAP + item_list[j]
    item_list[j] = "\n" + item_list[j]



#!/usr/bin/env python
# -*- coding:utf-8 -*-
faker_config = {
    'int': {
        'date_time_this_month': [
            'date', 'time'
        ]
    },
    'unicode': {
        'country_code': [
            'country_code'
        ],
        'address': [
            'address', 'street'
        ],
        'postcode': [
            'postcode'
        ],
        'zipcode': [
            'zipcode'
        ],
        'city': [
            'city'
        ],
        'country': [
            'country'
        ],
        'ean8': [
            'code', '_no', '_number'
        ],
        'company': [
            'company'
        ],
        'currency_code': [
            'currency'
        ],
        'free_email': [
            'email'
        ],
        'url': [
            'url', 'attachment'
        ],
        'job': [
            'job', 'role'
        ],
        'text': [
            'text', 'note'
        ],
        'name': [
            'name'
        ],
        'phone_number': [
            'phone'
        ],
        'uuid': [
            'uuid'
        ],
        'object_id': [
            'id'
        ]
    }
}
